from datetime import datetime, date, timedelta

def _fetch_assertions(top_left=None,
    bottom_right=None,
    coords=None,
    start_date: datetime = date.today()-timedelta(days=365),
    end_date: datetime = date.today(),
    ascending: bool = True,
    scale: int = 20,
    n_jobs: int = 8):

    assert(coords is None or (
        (
            type(coords) == list
            or type(coords) == tuple
        )
        and len(coords) == 2)
        and len(coords[0]) == len(coords[1])
        and len(coords[0]) == 2
    )
    assert(
            (
                top_left is None
                and bottom_right is None
            )
            or (
                type(top_left) == type(bottom_right)
                and (
                    type(top_left) == tuple
                    or type(top_left) == list)
            )
            and len(top_left) == len(bottom_right)
            and len(top_left) == 2
    )
    assert(start_date is not None)
    assert(end_date is not None)
    assert(end_date > start_date)

    if (top_left is not None
            and bottom_right is not None
            and coords is not None
    ):
        raise ValueError(
            "coords must be None if top_left and bottom_right are not None.")

def _fetch_point_assertions(
    coords,
    start_date: datetime = date.today()-timedelta(days=365),
    end_date: datetime = date.today(),
    ascending: bool = True,
    scale: int = 20,
    n_jobs: int = 8):

    assert(len(coords) == 2 and type(coords[0]) != list and type(coords[0]) != tuple)
    assert(start_date is not None)
    assert(end_date is not None)
    assert(end_date > start_date)


# LIBRARY IMPORTS
import ee
import warnings
from datetime import datetime, date, timedelta
from tqdm import tqdm
from functools import cmp_to_key
import numpy as np
from joblib import Parallel, delayed
import os

# LOCAL IMPORTS
from .constants import ASCENDING, DESCENDING
from .coordinates import populate_coordinates_dictionary
from .exceptions import IncorrectOrbitException
from .filter import filter_sentinel1_data
from .fetcher import fetch_sentinel1_data
from .fetcher import _get_zone_between_dates
from .utils import make_polygon
from .utils import tile_coordinates
from .utils import define_image_shape
from .utils import retrieve_max_pixel_count_from_pattern
from .utils import cmp_coords
from .utils import get_date_interval_array

def get_pixel_values(top_left=None, bottom_right=None, coords=None, start_date: datetime = date.today()-timedelta(days=365), end_date: datetime = date.today(), ascending: bool = True, scale=20, n_jobs=1):
    '''Given a list of coordinates and another of date intervals, loops over both to agglomerate GEE sar pixel values and returns a list of dictionnaries, one per coordinate, with intesities and timestamps.
    The higher loop is over coordinates and is concurrent. The lower loop is over dates and is parallelised.

    Parameters
    ----------
    top_left : tuple of float, optional
        Top left coordinates (lon, lat) of the Region 

    bottom_right : tuple of float, optional
        Bottom right coordinates (lon, lat) of the Region

    coords : tuple of tuple of float or list of list of float, optional
        If `top_left` and `bottom_right` are not specified, we expect `coords`
        to be a list (resp. tuple) of the form ``[top_left, bottom_right]``
        (resp. ``(top_left, bottom_right)``)

    start_date: datetime.datetime
        First date of the time interval

    end_date: datetime.datetime
        Last date of the time interval

    ascending : boolean, optional
        The trajectory to use when selecting data

    scale : int, optional
        Scale parameters of the getRegion() function. Defaulting at ``20``,
        change it to change the scale of the final data points. The highest,
        the lower the spatial resolution. Should be at least ``10``.

    n_jobs: int
        Defines the number of threads used for the parallelisation of date acquisitions

    ascending: bool

    Returns
    -------
    `list`
        A list of dictionnaries, each providing data over one given coordinate.
    '''

    orbit = ASCENDING if ascending else DESCENDING

    if (top_left is not None):
        list_of_coordinates = [make_polygon(top_left, bottom_right)]
    else:
        list_of_coordinates = [coords]

    date_intervals = get_date_interval_array(start_date, end_date)

    # retrieving the number of pixels per image
    try:
        polygon = ee.Geometry.Polygon(list_of_coordinates)
        _ = fetch_sentinel1_data(
            start_date=date_intervals[0][0],
            end_date=date_intervals[-1][1],
            geometry=polygon,
            scale=scale,
            orbit=orbit,
        )
    except Exception as e:
        # If the area is found to be too big
        if (str(e) == "ImageCollection.getRegion: No bands in collection."):
            raise IncorrectOrbitException(orbit)
        total_count_of_pixels = retrieve_max_pixel_count_from_pattern(str(e))
        if top_left is not None:
            list_of_coordinates = tile_coordinates(
                total_count_of_pixels, (top_left, bottom_right))
        else:
            list_of_coordinates = tile_coordinates(
                total_count_of_pixels, coords)


    print(f"Region sliced in {len(list_of_coordinates)} subregions and {len(date_intervals)} time intervals.")

    per_coord_dict = {}
    for c in tqdm(list_of_coordinates):
        vals = []
        headers = []
        polygon = ee.Geometry.Polygon([
            c
        ])
        # Updates vals and headers, one by aggregating returned values of the delayed function, the other by modifying the passed `headers` argument
        vals = Parallel(n_jobs=n_jobs, require="sharedmem")(delayed(_get_zone_between_dates)(sub_start_date, sub_end_date, polygon, scale, orbit, headers) for sub_start_date, sub_end_date in date_intervals)
        vals = [val for val in vals if val is not None]
        dictified_vals = [dict(zip(headers, val)) for values in vals for val in values]
        per_coord_dict = populate_coordinates_dictionary(
                dictified_values=dictified_vals,
                coordinates_dictionary=per_coord_dict,
        )
    pixel_values = [per_coord_dict[k] for k in per_coord_dict.keys()]
    cmp_coordinates = cmp_to_key(cmp_coords)
    pixel_values.sort(key=cmp_coordinates)  # sorting pixels by latitude then longitude
    
    return pixel_values

def get_point_pixel_values(coords, start_date: datetime = date.today()-timedelta(days=365), end_date: datetime = date.today(), ascending: bool = True, scale=20, n_jobs=1):
    '''Given a coordinate tuple and a list of date intervals, loops over both to agglomerate GEE sar pixel values and returns a list of dictionnaries, with intesities and timestamps.
    The main loop, over dates, is parallelised.


    Parameters
    ----------
    coords : tuple of float
        Coordinates (lon, lat) of the point of interest 
    
    start_date : datetime.datetime, optional
        First date of the time interval

    end_date : datetime.datetime, optional
        Last date of the time interval

    ascending : boolean, optional
        The trajectory to use when selecting data

    scale : int, optional
        Scale parameters of the getRegion() function. Defaulting at ``20``,
        change it to change the scale of the final data points. The highest,
        the lower the spatial resolution. Should be at least ``10``.

    n_jobs : int, optional
        Set the parallelisation factor (number of threads) for the GEE data
        access process. Set to 1 if no parallelisation required.

    Returns
    -------
    `list`
        A list with a single dictionnary, to match the behaviour of :func:`get_pixel_values` for a single point.
    '''
    orbit = ASCENDING if ascending else DESCENDING
    date_intervals = get_date_interval_array(start_date, end_date)
    polygon = ee.Geometry.Point(coords)
    headers = []
    # Updates vals and headers, one by aggregating returned values of the delayed function, the other by modifying the passed `headers` argument
    vals = Parallel(n_jobs=n_jobs, require="sharedmem")(delayed(_get_zone_between_dates)(sub_start_date, sub_end_date, polygon, scale, orbit, headers) for sub_start_date, sub_end_date in tqdm(date_intervals))
    vals = [val for val in vals if val is not None]
    dictified_vals = [dict(zip(headers, val)) for values in vals for val in values]
    per_coord_dict = populate_coordinates_dictionary(
        dictified_values=dictified_vals,
        coordinates_dictionary={}
    )

    pixel_values = [per_coord_dict[k] for k in per_coord_dict.keys()]
    cmp_coordinates = cmp_to_key(cmp_coords)
    pixel_values.sort(key=cmp_coordinates)  # sorting pixels by latitude then longitude
    
    return pixel_values

def get_timestamps_from_pixel_values(pixel_values):
    '''Given a pixel_values object from :func:`get_pixel_values`, retrieves a list of each unique acquisition days

    Parameters
    ----------
    pixel_values: list


    '''

    timestamps = np.unique(
            [
                datetime.fromtimestamp(pixel_values[i]['timestamps'][j]).date()
                for i in range(len(pixel_values))
                for j in range(len(pixel_values[i]['timestamps']))
            ]
    )

    return timestamps

def generate_image(timestamps, pixel_values):
    '''Given a list of timestamps and an array of pixel_values as :func:`~get_pixel_values`, generate a numpy matrix. Missing coordinates/dates will be inplaced with NaNs.

    Parameters
    ----------
    timestamps: list, or numpy.ndarray
        A list of unique dates (unicity is respected by measures of day: i.e, two acquisitions on the same day will be averaged)
    
    pixel_values: list
        A list of dictionnaries, each providing data over one given coordinate, preferably returned by :func:`~get_pixel_values`

    Returns
    -------
    `tuple`
        Returns a tuple containing the 4-D image as a `numpy.ndarray` of shape ``(height, width, polarisations, time)`` and a coordinates matrix, affecting to each pixel of the image its respective coordinate, of shape ``(height, width, 2)``.

    '''

    # Creating matrix of coordinates
    lats, lons = tuple(zip(*[(p["lat"], p["lon"]) for p in pixel_values]))
    
    unique_lats = np.unique(lats)
    unique_lats = unique_lats[::-1]
    lats_dict = {unique_lats[i]: i for i in range(len(unique_lats))}

    unique_lons = np.unique(lons)
    lons_dict = {unique_lons[i]: i for i in range(len(unique_lons))}

    width, height = len(unique_lons), len(unique_lats)
    coordinates = [[lat, lon] for lon in unique_lons for lat in unique_lats]
    coordinates = np.array(coordinates).reshape(height, width, 2)

    img = np.full((height, width, 2, len(timestamps)),fill_value=np.nan)

    print(f"Generating image of shape {height, width}")
    for p in tqdm(pixel_values):
        x, y = lats_dict[p["lat"]], lons_dict[p["lon"]]
        vv = []
        vh = []
        for timestamp in timestamps:

            indexes = np.argwhere(
                np.array([datetime.fromtimestamp(p_t).date() for p_t in p["timestamps"]]) == timestamp
            )
            vv.append(np.nanmean(
                np.array(p["VV"], dtype=float)[indexes]))
            vh.append(np.nanmean(
                np.array(p["VH"], dtype=float)[indexes]))

        img[x, y, 0, :] = vv
        img[x, y, 1, :] = vh
    return img, coordinates


def populate_coordinates_dictionary(dictified_values, coordinates_dictionary={}):
    """
    The dictionnary coordinates_dictionary' will be populated (or updated) with
    values from the 'dictified_values' dictionnary

    Parameters
    ----------
    dictified_values : dict
        A dictionary of Sentinel-1 values

    coordinates_dictionary : dict
        A dictionnary matching to each coordinate key its values through time
        as well as its timestamps.

    Returns
    -------
    coordinates_dictionary
    """
    for entry in dictified_values:
        lat = entry["latitude"]
        lon = entry["longitude"]
        new_key = str(lat)+":"+str(lon)

        if new_key in coordinates_dictionary:
            # Retrieving measured value
            coordinates_dictionary[new_key]["VV"].append(entry["VV"])
            coordinates_dictionary[new_key]["VH"].append(entry["VH"])
            tmstp = entry["time"]
            coordinates_dictionary[new_key]["timestamps"].append(tmstp//1000)

        else:
            coordinates_dictionary[new_key] = {}
            # Retrieving measured value
            coordinates_dictionary[new_key]["lat"] = lat
            coordinates_dictionary[new_key]["lon"] = lon
            coordinates_dictionary[new_key]["VV"] = [entry["VV"]]
            coordinates_dictionary[new_key]["VH"] = [entry["VH"]]
            tmstp = entry["time"]
            coordinates_dictionary[new_key]["timestamps"] = [tmstp//1000]

    return coordinates_dictionary



from geesarfetcher.filter import filter_sentinel1_data
from geesarfetcher.constants import VV, VH, IW, ASCENDING

__all__ = ["fetch_sentinel1_data"]

def fetch_sentinel1_data(start_date, end_date, geometry, scale, orbit=ASCENDING):
    '''
    Retrieves and queries ImageCollection using input parameters and return data as a tuple of header and values.

    Parameters
    ----------
    start_date : str
        str following the pattern ``'yyyy-mm-dd'`` describing the start date of the time interval
    end_date : str
        str following the pattern ``'yyyy-mm-dd'`` describing the end date of the time interval
    geometry : ee.Geometry
        Geometry object defining the area of process
    scale : int
        Scale parameters of the getRegion() function. Defaulting at ``20``, change it to change the scale of the final data points. The highest, the lower the spatial resolution. Should be at least ``10``.
    orbit : str, optional
        Defines the orbit to set for the data retrieval process

    Returns
    -------
    (val_header, val) : tuple
        val_header corresponds to the ``list of str`` describing the fields of the val array. The val array is a ``list`` of data records, each represented as a ``list`` of the same size as the val_header array.

    '''
    sentinel_1_roi = filter_sentinel1_data(
            start_date=start_date,
            end_date=end_date,
            geometry=geometry,
            orbit=orbit
    )
    val_vv = (sentinel_1_roi
              .select(VV)
              .getRegion(geometry, scale=scale)
              .getInfo()
    )
    val_vh = (sentinel_1_roi
              .select(VH)
              .getRegion(geometry, scale=scale)
              .getInfo()
    )
    val_header = val_vv[0][1:] + [VH]
    val = [
            val_vv[i][1:] + [val_vh[i][val_vh[0].index(VH)]] for i in range(1, len(val_vv))
    ]
    return (val_header, val)



def _get_zone_between_dates(start_date, end_date, polygon, scale, orbit, headers):
    try:
        val_header, val = fetch_sentinel1_data(
            start_date=start_date,
            end_date=end_date,
            geometry=polygon,
            orbit=orbit,
            scale=scale,
        )

        if len(headers) == 0:
            headers.extend(val_header)
        
        return val

    except Exception as e:
        pass


import math
from datetime import timedelta

MAX_GEE_PIXELS_DOWNLOAD = 1048576
GEE_ERROR_PLACEHOLDER = "ImageCollection.getRegion: Too many values: "

__all__ = ('tile_coordinates', 'retrieve_max_pixel_count_from_pattern',
           'cmp_coords', 'get_date_interval_array', 'make_polygon')


def make_polygon(top_left, bottom_right):
    '''Given two (lon, lat) coordinates of both the top left and bottom right corner of a polygon, return the list of corner coordinates of this polygon

    Parameters
    ----------
    top_left : list of int or tuple of int
        Top Left coordinates of the polygon
    bottom_right : list of int or tuple of int
        Bottom right coordinates of the polygon

    Returns
    -------
    list
        2-D list of the 5 coordinates need to create a Rectangular Polygon ``[top_left, top_right, bottom_right, bottom_left, top_left]``.
    '''
    return [
        list(top_left),
        [bottom_right[0], top_left[1]],
        list(bottom_right),
        [top_left[0], bottom_right[1]],
        list(top_left)
    ]


def tile_coordinates(total_count_of_pixels, coordinates, max_gee=MAX_GEE_PIXELS_DOWNLOAD):
    '''Given a coordinates array describing a Polygon, a count of pixes within that polygons, tiles this polygon into a grid a sub-Polygons where each sub-Polygon size matches the max_gee pixel count given as a parameter.

    Parameters
    ----------
    total_count_of_pixels : int
        Total number of pixels of the designated area
    coordinates : array of array of floats
        Can be a 5-sized list of every coordinates defining the polygon ``[[long1, lat1],[long2, lat1]...,[long1, lat1]]`` or a 2-sized list of coordinates defining the top left and bottom right corner of the Polygon ``[[long1, lat1],[long2, lat2]]``
    max_gee_threshold : int, optional
        Total number of points allowed for one data query. Default: 1048576

    Returns
    -------
    list
        3-dimensional list of coordinates with pixel count inferior or equal to the maximum GEE threshold (shape: ``(number of images, number of coordinates per image, 2)``)

    '''
    assert(len(coordinates) == 2 or len(coordinates) == 5)

    list_of_coordinates = []

    if len(coordinates) == 2:
        tmp_c = make_polygon(coordinates[0], coordinates[1])
    else:
        tmp_c = coordinates

    if (total_count_of_pixels < max_gee):
        return [tmp_c]

    # The coordinate polygon will be tiled in `grid_length * grid_length` sub-Polygons
    grid_length = int(math.ceil(math.sqrt(total_count_of_pixels/max_gee)))

    original_polygon_width = tmp_c[1][0] - tmp_c[0][0]
    original_polygon_height = tmp_c[3][1] - tmp_c[0][1]

    for i in range(grid_length):
        for j in range(grid_length):
            list_of_coordinates.append(
                make_polygon(
                    [tmp_c[0][0]+i*original_polygon_width/grid_length,
                        tmp_c[0][1]+j*original_polygon_height/grid_length], 
                    [tmp_c[0][0]+(i+1)*original_polygon_width/grid_length,
                        tmp_c[0][1]+(j+1)*original_polygon_height/grid_length]
                )
            )

    return list_of_coordinates


def retrieve_max_pixel_count_from_pattern(error_str):
    '''Given an input getRegion error from GEE, extract the provided points count.

    Parameters
    ----------
    error_str : str
        the str text of the GEE error (e.g. the function caled on ``"ImageCollection.getRegion: Too many values: x points ..."`` will output x)

    Returns
    -------
    int
        Returns the number of points specified in the input image
    '''
    try:
        return int(error_str.split("ImageCollection.getRegion: Too many values: ")[1].split(" points")[0])
    except:
        raise ValueError("No max pixels value found")


def cmp_coords(a, b):
    '''
    Given two coordinates dict a and b, compare which one is closer to the North-Eastern direction

    Parameters
    ----------
    a : dict 
        dict with keys ``"lon"`` and ``"lat"``
    b : dict
        dict with keys ``"lon"`` and ``"lat"``

    Returns
    -------
    int
        **-1** if ``a > b``, **1** if ``a < b``, **0** if ``a == b``
    '''
    if a["lat"] != b["lat"]:
        return 1 if a["lat"] < b["lat"] else -1
    elif a["lon"] != b["lon"]:
        return 1 if a["lon"] < b["lon"] else -1
    else:
        return 0


def define_image_shape(pixel_values):
    """Define image shape based on number pixel and latitude values

    Parameters
    ----------
    pixel_values :
        Dictionnary with retrieved pixel values along with latitude and
        longitude coordinates

    Returns
    -------
    (with, height) :
        A tuple with the width and height of the requested area of interest
    """
    # count pixels with common latitude until it changes to know the image width
    width = 1
    while pixel_values[width]["lat"] == pixel_values[0]["lat"]:
        width += 1
    # deduce the image height from its width
    height = len(pixel_values) // width
    return (width, height)


def get_date_interval_array(start_date, end_date, day_timedelta=1):
    '''Initialize a list of days interval of size ``day_timedelta`` iteratively created between ``start_date`` and ``end_date``.

    Parameters
    ----------
    start_date : datetime.datetime
        first date time of the array
    end_date : datetime.datetime
        last date of the array
    day_timedelta : int
        size, in days, of every interval
    '''
    assert(start_date is not None and end_date is not None and day_timedelta is not None)
    assert(start_date < end_date)
    assert(type(day_timedelta) == int)

    date_intervals = []

    tmp_date = start_date

    while tmp_date < end_date:
        date_intervals.append((tmp_date.strftime(
            "%Y-%m-%d"), (tmp_date + timedelta(days=day_timedelta)).strftime("%Y-%m-%d")))
        tmp_date += timedelta(days=day_timedelta)

    return date_intervals


from geesarfetcher import fetch, fetch_point
import unittest
import numpy as np
from datetime import datetime, date, timedelta
import warnings
warnings.simplefilter(action="ignore")

# SMALL COORDS
SMALL_COORDS = [
    [-104.71539380321867, 41.77751907620989],
    [-104.7079050756491, 41.77751907620989],
    [-104.7079050756491, 41.78321546046257],
    [-104.71539380321867, 41.78321546046257],
    [-104.71539380321867, 41.77751907620989]
]

MED_COORDS = [
    [-104.77431630331856, 41.729889598264826],
    [-104.65140675742012, 41.729889598264826],
    [-104.65140675742012, 41.81515375846025],
    [-104.77431630331856, 41.81515375846025],
    [-104.77431630331856, 41.729889598264826]
]

BIG_COORDS = [
    [-104.88572453696113, 41.62148183942426],
    [-104.53690861899238, 41.62148183942426],
    [-104.53690861899238, 41.884778748257574],
    [-104.88572453696113, 41.884778748257574],
    [-104.88572453696113, 41.62148183942426]
]


class TestFetcher(unittest.TestCase):
    def test_small_download(self):
        '''Tests download with a small region and small time interval
        '''
        d = fetch(
            top_left=SMALL_COORDS[0],
            bottom_right=SMALL_COORDS[2],
            start_date=datetime(year=2020, month=10, day=24),
            end_date=datetime(year=2020, month=11, day=2),
            ascending=False,
            scale=10)

        self.assertTrue(isinstance(d, dict))
        self.assertTrue("timestamps" in list(d.keys())
                        and len(d["timestamps"]) == 2)
        self.assertTrue(d["stack"].shape == (64, 83, 2, 2))

    def test_medium_download(self):
        '''Tests download with a medium region (at least one cut) and small time interval
        '''
        d = fetch(
            top_left=MED_COORDS[0],
            bottom_right=MED_COORDS[2],
            start_date=datetime(year=2020, month=10, day=24),
            end_date=datetime(year=2020, month=11, day=2),
            ascending=False,
            scale=10)

        self.assertTrue(isinstance(d, dict))
        self.assertTrue("timestamps" in list(d.keys())
                        and len(d["timestamps"]) == 2)
        self.assertTrue(d["stack"].shape == (949, 1368, 2, 2))

    def test_big_scale_download(self):
        '''Tests download with a large scale and missing pixel values
        '''
        d = fetch(
            top_left = [5.26948161, 49.97440374],
            bottom_right = [8.07913236, 47.90621773],
            start_date = datetime(2019, 6, 1),
            end_date = datetime(2019, 6, 3),
            ascending=False,
            scale=1000,
        )
        self.assertEqual(d["stack"].shape, (231, 312, 2, 1))
        self.assertEqual(d["stack"][100,100,0,0], -11.805720298590167)

    def test_fetch_point_small_scale(self):
        '''Tests fetch_point method with a relative small scale of 10
        '''
        d = fetch_point(
            coords = [-104.88572453696113, 41.884778748257574],
            start_date=datetime(year=2020, month=10, day=24),
            end_date=datetime(year=2020, month=11, day=2),
            ascending=False,
            scale=10
        )
        self.assertTrue(d['stack'].shape == (1,1,2,2))
        self.assertTrue(np.round(d['stack'][0,0,0,0] - (-15.11937733), 2) == 0.0)

    def test_fetch_point_med_scale(self):
        '''Tests fetch_point method with a relative medium scale of 100
        '''
        d = fetch_point(
            coords = [-104.88572453696113, 41.884778748257574],
            start_date=datetime(year=2020, month=10, day=24),
            end_date=datetime(year=2020, month=11, day=2),
            ascending=False,
            scale=100
        )
        self.assertTrue(d['stack'].shape == (1,1,2,2))
        self.assertTrue(np.round(d['stack'][0,0,0,0] - (-15.04562971), 2) == 0.0)

    def test_fetch_point_med_scale(self):
        '''Tests fetch_point method with a relative medium scale of 100
        '''
        d = fetch_point(
            coords = [-104.88572453696113, 41.884778748257574],
            start_date=datetime(year=2020, month=10, day=24),
            end_date=datetime(year=2020, month=11, day=2),
            ascending=False,
            scale=1000
        )
        self.assertTrue(d['stack'].shape == (1,1,2,2))
        self.assertTrue(np.round(d['stack'][0,0,0,0] - (-13.21006906), 2) == 0.0)

    def test_exceptions(self):
        '''Tests raised exceptions
        '''
        # Redundant information provided as input
        kwargs = {
            "top_left": MED_COORDS[0],
            "bottom_right": MED_COORDS[2],
            "coords": [MED_COORDS[0],MED_COORDS[2]],
            "start_date": datetime(year=2020, month=10, day=24),
            "end_date": datetime(year=2020, month=11, day=2),
            "ascending": False,
            "scale": 10
        }
        self.assertRaises(ValueError, fetch, **kwargs)

        # No coordinates provided
        kwargs["top_left"] = kwargs["bottom_right"] = kwargs["coords"] = None
        self.assertRaises(ValueError, fetch, **kwargs)

        # Only one coordinate provided
        kwargs["top_left"] = [0.0, 0.0]
        self.assertRaises(AssertionError, fetch, **kwargs)

        # Missing start date provided
        kwargs["top_left"] = MED_COORDS[0]
        kwargs["bottom_right"] = MED_COORDS[2]
        kwargs["start_date"] = None
        self.assertRaises(AssertionError, fetch, **kwargs)

        # Missing end date provided
        kwargs["start_date"] = datetime(year=2020, month=8, day=16)
        kwargs["end_date"] = None
        self.assertRaises(AssertionError, fetch, **kwargs)
        



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
from setuptools import setup, find_packages
import itertools
# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))

# Parse the version from the geextract module.
with open('geesarfetcher/__init__.py') as f:
    for line in f:
        if line.find("__version__") >= 0:
            version = line.split("=")[1].strip()
            version = version.strip('"')
            version = version.strip("'")
            continue

# 
extra_reqs = {'docs': ['sphinx',
                       'sphinx-rtd-theme',
                       'sphinxcontrib-programoutput']}
extra_reqs['all'] = list(set(itertools.chain(*extra_reqs.values())))

with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    readme = f.read()
    
setup(name='geesarfetcher',
      version=version,
      description=u"Extract Sentinel-1 GRD time-series images over a given area from google earth engine",
      long_description=readme,
      long_description_content_type="text/markdown",
      classifiers=[],
      keywords='SAR, sentinel, google, gee, time-series, radar, satellite imagery',
      author=u"Thomas Di Martino",
      author_email='thomas.di-martino@hotmail.com',
      url='https://github.com/dimartinot/gee_sar_fetcher',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          'tqdm>=4.50.2',
          'numpy>=1.18.0,<=1.19.3',
          'earthengine-api>=0.1.231',
          'joblib>=0.17.0'],
      #scripts=['geextract/scripts/gee_extract.py',
      #         'geextract/scripts/gee_extract_batch.py'],
      test_suite="tests",
      extras_require=extra_reqs)



class IncorrectOrbitException(Exception):
    '''Exception class triggered when the specified orbit does not retrieve any data between the two passed dates
    '''
    def __init__(self, orbit):
        msg = f"Incorrect orbit '{orbit}': No bands found in collection, please visit https://sentinel.esa.int/web/sentinel/missions/sentinel-1/observation-scenario for more info."
        super(IncorrectOrbitException, self).__init__(msg)


from geesarfetcher.constants import SENTINEL1_COLLECTION_ID
from geesarfetcher.constants import VV, VH, IW
from geesarfetcher.constants import ASCENDING
import ee

def filter_sentinel1_data(start_date, end_date, geometry, orbit=ASCENDING):
    '''Filters Sentinel-1 products to get images collected in interferometric
    wide swath mode (IW) and on i) a date range, ii) a geometry and iii)
    ascending or descending orbit.

    Parameters
    ----------
    start_date : str
        str following the pattern ``'yyyy-mm-dd'`` describing the start date of the time interval
    end_date : str
        str following the pattern ``'yyyy-mm-dd'`` describing the end date of the time interval
    geometry : ee.Geometry
        Geometry object defining the area of process
    orbit : str, optional
        Defines the orbit to set for the data retrieval process

    Returns:
    --------
    ee.ImageCollection
        Filtered ImageCollection left to be queried
    '''
    filtered_data = (ee.ImageCollection(SENTINEL1_COLLECTION_ID)
                      .filter(ee.Filter.date(start_date, end_date))
                      .filterBounds(geometry)
                      .filter(ee.Filter.listContains('transmitterReceiverPolarisation', VV))
                      .filter(ee.Filter.listContains('transmitterReceiverPolarisation', VH))
                      .filter(ee.Filter.eq('instrumentMode', IW))
                      .filter(ee.Filter.eq('orbitProperties_pass', orbit))
                      .filter(ee.Filter.eq('resolution_meters', 10))
    )
    return filtered_data



__all__ = ['SENTINEL1_COLLECTION_ID', 'VV', 'VH', 'IW', 'ASCENDING', 'DESCENDING']

SENTINEL1_COLLECTION_ID = 'COPERNICUS/S1_GRD'
VV = 'VV'
VH = 'VH'
IW = 'IW'
ASCENDING = 'ASCENDING'
DESCENDING = 'DESCENDING'



"""geesarfetcher"""

__version__ = "0.3.0"

# LIBRARY IMPORTS
import ee
import warnings
from datetime import datetime, date, timedelta
import os

# LOCAL IMPORTS
from .api import generate_image
from .api import get_pixel_values
from .api import get_point_pixel_values
from .api import get_timestamps_from_pixel_values
from .assertions import _fetch_assertions
from .assertions import _fetch_point_assertions


warnings.simplefilter(action="ignore")
warnings.filterwarnings('ignore')

if os.environ.get('READTHEDOCS') == None:
    ee.Initialize()

def fetch(
    top_left=None,
    bottom_right=None,
    coords=None,
    start_date: datetime = date.today()-timedelta(days=365),
    end_date: datetime = date.today(),
    ascending: bool = True,
    scale: int = 20,
    n_jobs: int = 8
):
    '''Fetches SAR data in the form of a dictionnary with image data as well as timestamps

    Parameters
    ----------
    top_left : tuple of float, optional
        Top left coordinates (lon, lat) of the Region 

    bottom_right : tuple of float, optional
        Bottom right coordinates (lon, lat) of the Region

    coords : tuple of tuple of float or list of list of float, optional
        If `top_left` and `bottom_right` are not specified, we expect `coords`
        to be a list (resp. tuple) of the form ``[top_left, bottom_right]``
        (resp. ``(top_left, bottom_right)``)

    start_date : datetime.datetime, optional
        First date of the time interval

    end_date : datetime.datetime, optional
        Last date of the time interval

    ascending : boolean, optional
        The trajectory to use when selecting data

    scale : int, optional
        Scale parameters of the getRegion() function. Defaulting at ``20``,
        change it to change the scale of the final data points. The highest,
        the lower the spatial resolution. Should be at least ``10``.

    n_jobs : int, optional
        Set the parallelisation factor (number of threads) for the GEE data
        access process. Set to 1 if no parallelisation required.

    Returns
    -------
    `dict`
        Dictionnary with four keys:

            ``"stacks"``
                4-D array containing db intensity measure (`numpy.ndarray`),
                ``(height, width, pol_count, time_series_length)`` 

            ``"coordinates"``
                3-D array containg coordinates where ``[:,:,0]`` provides
                access to latitude and ``[:,:,1]`` provides access to
                longitude, (`numpy.ndarray`), ``(height, width, 2)``

            ``"timestamps"``
                list of acquisition timestamps of size (time_series_length,)
                (`list of str`)

            ``"metadata"``
                Dictionnary describing data for each axis of the stack and the
                coordinates
    '''

    _fetch_assertions(top_left, bottom_right, coords, start_date, end_date, ascending, scale, n_jobs)    

    pixel_values = get_pixel_values(top_left=top_left, bottom_right=bottom_right, coords=coords, start_date=start_date, end_date=end_date, ascending=ascending, scale=scale, n_jobs=n_jobs)
    timestamps = get_timestamps_from_pixel_values(pixel_values)
    img, coordinates = generate_image(timestamps, pixel_values)

    return {
        "stack": img,
        "timestamps": timestamps,
        "coordinates": coordinates,
        "metadata": {
            "stack": {
                "axis_0": "height",
                "axis_1": "width",
                "axis_2": "polarisations (0:VV, 1:VH)",
                "axis_3": "timestamps"
            },
            "coordinates": {
                "axis_0": "height",
                "axis_1": "width",
                "axis_2": "0:latitude; 1:longitude",
            }
        }
    }

def fetch_point(
    coords,
    start_date: datetime = date.today()-timedelta(days=365),
    end_date: datetime = date.today(),
    ascending: bool = True,
    scale: int = 20,
    n_jobs: int = 8
):
    '''Fetches SAR data from a single coordinate point in the form of a dictionnary with image data as well as timestamps

    Parameters
    ----------
    coords : tuple of float
        Coordinates (lon, lat) of the point of interest 
    
    start_date : datetime.datetime, optional
        First date of the time interval

    end_date : datetime.datetime, optional
        Last date of the time interval

    ascending : boolean, optional
        The trajectory to use when selecting data

    scale : int, optional
        Scale parameters of the getRegion() function. Defaulting at ``20``,
        change it to change the scale of the final data points. The highest,
        the lower the spatial resolution. Should be at least ``10``.

    n_jobs : int, optional
        Set the parallelisation factor (number of threads) for the GEE data
        access process. Set to 1 if no parallelisation required.

    Returns
    -------
    `dict`
        Dictionnary with four keys:

            ``"stacks"``
                4-D array containing db intensity measure (`numpy.ndarray`),
               ``(1, 1, pol_count, time_series_length)`` 

            ``"coordinates"``
                3-D array containg coordinates where ``[:,:,0]`` provides
                access to latitude and ``[:,:,1]`` provides access to
                longitude, (`numpy.ndarray`), ``(1, 1, 2)``

            ``"timestamps"``
                list of acquisition timestamps of size (time_series_length,)
                (`list of str`)

            ``"metadata"``
                Dictionnary describing data for each axis of the stack and the
                coordinates

    '''

    _fetch_point_assertions(coords, start_date, end_date, ascending, scale, n_jobs)

    pixel_values = get_point_pixel_values(coords=coords, start_date=start_date, end_date=end_date, ascending=ascending, scale=scale, n_jobs=n_jobs)
    timestamps = get_timestamps_from_pixel_values(pixel_values)
    img, coordinates = generate_image(timestamps, pixel_values)

    return {
        "stack": img,
        "timestamps": timestamps,
        "coordinates": coordinates,
        "metadata": {
            "stack": {
                "axis_0": "height",
                "axis_1": "width",
                "axis_2": "polarisations (0:VV, 1:VH)",
                "axis_3": "timestamps"
            },
            "coordinates": {
                "axis_0": "height",
                "axis_1": "width",
                "axis_2": "0:latitude; 1:longitude",
            }
        }
    }